@@ -28,6 +28,7 @@
 
 #include "fastmath_p.h"
 #include "microexif_p.h"
+#include "packbits_p.h"
 #include "psd_p.h"
 #include "scanlineconverter_p.h"
 #include "util_p.h"
@@ -71,6 +72,16 @@ typedef quint8 uchar;
  */
 // #define PSD_FORCE_RGBA
 
+/* *** PSD_MAX_IMAGE_WIDTH and PSD_MAX_IMAGE_HEIGHT ***
+ * The maximum size in pixel allowed by the plugin.
+ */
+#ifndef PSD_MAX_IMAGE_WIDTH
+#define PSD_MAX_IMAGE_WIDTH KIF_LARGE_IMAGE_PIXEL_LIMIT
+#endif
+#ifndef PSD_MAX_IMAGE_HEIGHT
+#define PSD_MAX_IMAGE_HEIGHT PSD_MAX_IMAGE_WIDTH
+#endif
+
 namespace // Private.
 {
 
@@ -545,7 +556,7 @@ static bool setExifData(QImage &img, const MicroExif &exif)
 {
     if (exif.isEmpty())
         return false;
-    exif.toImageMetadata(img);
+    exif.updateImageMetadata(img);
     return true;
 }
 
@@ -677,7 +688,7 @@ static bool IsValid(const PSDHeader &header)
         qDebug() << "PSD header: invalid number of channels" << header.channel_count;
         return false;
     }
-    if (header.width > 300000 || header.height > 300000) {
+    if (header.width > std::min(300000, PSD_MAX_IMAGE_WIDTH) || header.height > std::min(300000, PSD_MAX_IMAGE_HEIGHT)) {
         qDebug() << "PSD header: invalid image size" << header.width << "x" << header.height;
         return false;
     }
@@ -712,48 +723,6 @@ static bool IsSupported(const PSDHeader &header)
     return true;
 }
 
-/*!
- * \brief decompress
- * Fast PackBits decompression.
- * \param input The compressed input buffer.
- * \param ilen The input buffer size.
- * \param output The uncompressed target buffer.
- * \param olen The target buffer size.
- * \return The number of valid bytes in the target buffer.
- */
-qint64 decompress(const char *input, qint64 ilen, char *output, qint64 olen)
-{
-    qint64  j = 0;
-    for (qint64 ip = 0, rr = 0, available = olen; j < olen && ip < ilen; available = olen - j) {
-        signed char n = static_cast<signed char>(input[ip++]);
-        if (n == -128)
-            continue;
-
-        if (n >= 0) {
-            rr = qint64(n) + 1;
-            if (available < rr) {
-                --ip;
-                break;
-            }
-
-            if (ip + rr > ilen)
-                return -1;
-            memcpy(output + j, input + ip, size_t(rr));
-            ip += rr;
-        } else if (ip < ilen) {
-            rr = qint64(1-n);
-            if (available < rr) {
-                --ip;
-                break;
-            }
-            memset(output + j, input[ip++], size_t(rr));
-        }
-
-        j += rr;
-    }
-    return j;
-}
-
 /*!
  * \brief imageFormat
  * \param header The PSD header.
@@ -1102,7 +1071,7 @@ bool readChannel(QByteArray &target, QDataStream &stream, quint32 compressedSize
         if (stream.readRawData(tmp.data(), tmp.size()) != tmp.size()) {
             return false;
         }
-        if (decompress(tmp.data(), tmp.size(), target.data(), target.size()) < 0) {
+        if (packbitsDecompress(tmp.data(), tmp.size(), target.data(), target.size()) < 0) {
             return false;
         }
     } else if (stream.readRawData(target.data(), target.size()) != target.size()) {
@@ -1504,7 +1473,18 @@ bool PSDHandler::read(QImage *image)
         img.setColorSpace(QColorSpace(QColorSpace::SRgb));
 #endif
     } else if (!setColorSpace(img, irs)) {
-        // qDebug() << "No colorspace info set!";
+        // Float images are used by Photoshop as linear: if no color space
+        // is present, a linear one should be chosen.
+        if (header.color_mode == CM_RGB && header.depth == 32) {
+            img.setColorSpace(QColorSpace(QColorSpace::SRgbLinear));
+        }
+#if QT_VERSION >= QT_VERSION_CHECK(6, 8, 0)
+        if (header.color_mode == CM_GRAYSCALE && header.depth == 32) {
+            auto qs = QColorSpace(QPointF(0.3127, 0.3291), QColorSpace::TransferFunction::Linear);
+            qs.setDescription(QStringLiteral("Linear grayscale"));
+            img.setColorSpace(qs);
+        }
+#endif
     }
 
     // XMP data
