@@ -11,7 +11,6 @@
 
 #include <QColorSpace>
 #include <QDateTime>
-#include <QDebug>
 #include <QImage>
 #include <QSet>
 #include <QTimeZone>
@@ -22,12 +21,24 @@
 
 #include <libraw/libraw.h>
 
+#ifndef RAW_DISABLE_BROKEN_STREAM_PROTECTION
+/* The LibRaw_QIODevice stream interrupt the reading  of the RAW file if a
+ * burst of read errors are detected.
+ * By defining RAW_IGNORE_BROKEN_STREAMS removes this protection, leaving
+ * LibRaw in complete control of error handling (which is mostly absent).
+ */
+// #define RAW_DISABLE_BROKEN_STREAM_PROTECTION
+#endif
+
 #ifdef QT_DEBUG
-// This should be used to exclude the local QIODevice wrapper of the LibRaw_abstract_datastream interface.
-// If the result changes then the problem is in the local wrapper and must be corrected.
-// WARNING: using the LibRaw's streams instead of the local wrapper from a Qt program falls in the LOCALE
-//          bug when the RAW file needs the scanf_one() function (e.g. *.MOS files). See also raw_scanf_one().
-//#define EXCLUDE_LibRaw_QIODevice // Uncomment this code if you think that the problem is LibRaw_QIODevice (default commented)
+/* This should be used to exclude the local QIODevice wrapper of the
+ * LibRaw_abstract_datastream interface. If the result changes then the problem
+ * is in the local wrapper and must be corrected.
+ * WARNING: using the LibRaw's streams instead of the local wrapper from a Qt
+ *          program falls in the LOCALE bug when the RAW file needs the
+ *          scanf_one() function (e.g. *.MOS files). See also raw_scanf_one().
+ */
+// #define EXCLUDE_LibRaw_QIODevice // Uncomment this code if you think that the problem is LibRaw_QIODevice (default commented)
 #endif
 
 namespace // Private.
@@ -40,20 +51,18 @@ using pi_unique_ptr = std::unique_ptr<libraw_processed_image_t, std::function<vo
 // Known formats supported by LibRaw (in alphabetical order and lower case)
 const auto supported_formats = QSet<QByteArray>{
     "3fr",
-    "arw", "arq",
-    "bay", "bmq",
-    "cr2", "cr3", "cap", "cine", "cs1", "crw",
-    "dcs", "dc2", "dcr", "dng", "drf", "dxo",
-    "eip", "erf",
+    "arw",
+    "crw", "cr2", "cr3",
+    "dcr", "dng",
+    "erf",
     "fff",
     "iiq",
-    "k25", "kc2", "kdc",
-    "mdc", "mef", "mfw", "mos", "mrw",
+    "k25", "kdc",
+    "mdc", "mef", "mos", "mrw",
     "nef", "nrw",
-    "obm", "orf", "ori",
-    "pef", "ptx", "pxn",
-    "qtk",
-    "r3d", "raf", "raw", "rdc", "rw2", "rwl", "rwz",
+    "orf",
+    "pef",
+    "raf", "raw", "rwl", "rw2",
     "sr2", "srf", "srw", "sti",
     "x3f"
 };
@@ -97,10 +106,46 @@ inline int raw_scanf_one(const QByteArray &ba, const char *fmt, void *val)
  */
 class LibRaw_QIODevice : public LibRaw_abstract_datastream
 {
+    inline void incError()
+    {
+        // Normally, when read errors occur, a plugin exits with an error.
+        // Here, given how LibRaw works, we need to be a little more flexible:
+        // for example, a seek might fail when the library tries to figure out
+        // what it's reading. We therefore use a dynamic error counting system
+        // where the error counts heavily, but subsequent successful
+        // operations can mitigate it.
+        // If the stream is obviously corrupt, the plugin exits.
+        m_errorCount += 3;
+    }
+    inline void decError()
+    {
+        m_errorCount = std::max(m_errorCount - 1, 0);
+    }
+    void checkErrors() {
+        // Internally, LibRaw often doesn't check the return values of various
+        // functions. It's not uncommon to find things like:
+        //
+        //  fseek(ifp, oAtom, SEEK_SET);
+        //  szAtom = get4();
+        //
+        // which can cause the read to go into an infinite loop on corrupted
+        // files. I've set the error count to high to avoid these issues.
+        if (m_errorCount > 10) {
+#ifndef RAW_DISABLE_BROKEN_STREAM_PROTECTION
+            // LibRaw stream classes can throw exceptions that are handled by
+            // libraw itself (e.g. see implementation of
+            // LibRaw_bigfile_datastream).
+            // Every LibRaw function that works on the stream used in this
+            // plugin is protected by a Try...Catch.
+            IOERROR(); // defined in libraw_datastream.h
+#endif
+        }
+    }
 public:
     explicit LibRaw_QIODevice(QIODevice *device)
+        : m_device(device)
+        , m_errorCount(0)
     {
-        m_device = device;
     }
     virtual ~LibRaw_QIODevice() override
     {
@@ -111,21 +156,24 @@ public:
     }
     virtual int read(void *ptr, size_t sz, size_t nmemb) override
     {
+        checkErrors();
+
         qint64 read = 0;
         if (sz == 0) {
             return 0;
         }
         auto data = reinterpret_cast<char*>(ptr);
         for (qint64 r = 0, size = sz * nmemb; read < size; read += r) {
-            if (m_device->atEnd()) {
-                break;
-            }
             r = m_device->read(data + read, size - read);
             if (r < 1) {
+                if (read == 0 || r < 0) {
+                    incError();
+                }
                 break;
             }
+            decError();
         }
-        return read / sz;
+        return int(read / sz);
     }
     virtual int eof() override
     {
@@ -133,17 +181,21 @@ public:
     }
     virtual int seek(INT64 o, int whence) override
     {
+        checkErrors();
+
         auto pos = o;
-        auto size = m_device->size();
+        auto siz = size();
         if (whence == SEEK_CUR) {
-            pos = m_device->pos() + o;
+            pos = tell() + o;
         }
         if (whence == SEEK_END) {
-            pos = size + o;
+            pos = siz + o;
         }
-        if (pos < 0 || m_device->isSequential()) {
+        if (pos < 0 || pos > siz || m_device->isSequential()) {
+            incError();
             return -1;
         }
+        decError();
         return m_device->seek(pos) ? 0 : -1;
     }
     virtual INT64 tell() override
@@ -156,17 +208,24 @@ public:
     }
     virtual char *gets(char *s, int sz) override
     {
+        checkErrors();
+
         if (m_device->readLine(s, sz) > 0) {
+            decError();
             return s;
         }
+        incError();
         return nullptr;
     }
     virtual int scanf_one(const char *fmt, void *val) override
     {
+        checkErrors();
+
         QByteArray ba;
         for (int xcnt = 0; xcnt < 24 && !m_device->atEnd(); ++xcnt) {
             char c;
             if (!m_device->getChar(&c)) {
+                incError();
                 return EOF;
             }
             if (ba.isEmpty() && (c == ' ' || c == '\t')) {
@@ -177,14 +236,19 @@ public:
             }
             ba.append(c);
         }
+        decError();
         return raw_scanf_one(ba, fmt, val);
     }
     virtual int get_char() override
     {
+        checkErrors();
+
         unsigned char c;
         if (!m_device->getChar(reinterpret_cast<char *>(&c))) {
+            incError();
             return EOF;
         }
+        decError();
         return int(c);
     }
 #if (LIBRAW_VERSION < LIBRAW_MAKE_VERSION(0, 21, 0)) || defined(LIBRAW_OLD_VIDEO_SUPPORT)
@@ -196,6 +260,8 @@ public:
 
 private:
     QIODevice *m_device;
+
+    qint32 m_errorCount;
 };
 
 bool addTag(const QString &tag, QStringList &lines)
@@ -433,7 +499,7 @@ inline void rgbToRgbX(uchar *target, const uchar *source, qint32 targetSize, qin
 #define T_FLAGS(a) (((a) >> 31) & 0x1)
 // clang-format on
 
-#define DEFAULT_QUALITY (C_IQ(3) | C_OC(1) | C_CW(1) | C_AW(1) | C_BT(1) | C_HS(0) | C_FLAGS(1))
+#define DEFAULT_IMAGE_QUALITY (C_IQ(3) | C_OC(1) | C_CW(1) | C_AW(1) | C_BT(1) | C_HS(0) | C_FLAGS(1))
 
 void setParams(QImageIOHandler *handler, LibRaw *rawProcessor)
 {
@@ -497,7 +563,7 @@ void setParams(QImageIOHandler *handler, LibRaw *rawProcessor)
         quality |= C_FLAGS(1);
     }
     if (quality == -1) {
-        quality = DEFAULT_QUALITY;
+        quality = DEFAULT_IMAGE_QUALITY;
     }
     Q_ASSERT(T_FLAGS(quality));
 
@@ -581,6 +647,66 @@ void setParams(QImageIOHandler *handler, LibRaw *rawProcessor)
     params.use_fuji_rotate = T_SR(quality) ? 0 : 1;
 }
 
+bool LoadTHUMB(QImageIOHandler *handler, QImage &img)
+{
+    std::unique_ptr<LibRaw> rawProcessor(new LibRaw);
+
+    // *** Open the stream
+    auto device = handler->device();
+#ifndef EXCLUDE_LibRaw_QIODevice
+    LibRaw_QIODevice stream(device);
+    if (rawProcessor->open_datastream(&stream) != LIBRAW_SUCCESS) {
+        return false;
+    }
+#else
+    auto all = device->readAll();
+    if (rawProcessor->open_buffer(all.data(), all.size()) != LIBRAW_SUCCESS) {
+        return false;
+    }
+#endif
+
+#if (LIBRAW_VERSION < LIBRAW_MAKE_VERSION(0, 21, 0))
+    // *** Unpacking selected thumbnail
+    if (rawProcessor->unpack_thumb() != LIBRAW_SUCCESS) {
+        return false;
+    }
+#else
+    // *** Search for the bigger thumbnail
+    auto &&tlist = rawProcessor->imgdata.thumbs_list;
+    auto idx = 0;
+    for (auto n = 1; n < std::min(tlist.thumbcount, LIBRAW_THUMBNAIL_MAXCOUNT); ++n) {
+        if (tlist.thumblist[n].twidth > tlist.thumblist[idx].twidth)
+            idx = n;
+    }
+
+    // *** Unpacking selected thumbnail
+    if (rawProcessor->unpack_thumb_ex(idx) != LIBRAW_SUCCESS) {
+        return false;
+    }
+#endif
+
+    // *** Convert to QImage
+    pi_unique_ptr processedImage(rawProcessor->dcraw_make_mem_thumb(), LibRaw::dcraw_clear_mem);
+    if (processedImage == nullptr) {
+        return false;
+    }
+    auto ba = QByteArray(reinterpret_cast<const char *>(processedImage->data), qsizetype(processedImage->data_size));
+    if (ba.isEmpty()) {
+        return false;
+    }
+    if (processedImage->type == LIBRAW_IMAGE_BITMAP) {
+        // clang-format off
+        auto header = QString::fromUtf8("P%1\n%2 %3\n%4\n") // taken from KDcraw
+                          .arg(processedImage->colors == 3 ? QLatin1String("6") : QLatin1String("5"))
+                          .arg(processedImage->width)
+                          .arg(processedImage->height)
+                          .arg((1 << processedImage->bits)-1);
+        // clang-format on
+        ba.prepend(header.toLatin1());
+    }
+    return img.loadFromData(ba);
+}
+
 bool LoadRAW(QImageIOHandler *handler, QImage &img)
 {
     std::unique_ptr<LibRaw> rawProcessor(new LibRaw);
@@ -679,6 +805,11 @@ bool LoadRAW(QImageIOHandler *handler, QImage &img)
         if (params.output_color == 7) {
             img.setColorSpace(QColorSpace(QColorSpace::DisplayP3));
         }
+#if QT_VERSION >= QT_VERSION_CHECK(6, 8, 0)
+        if (params.output_color == 8) {
+            img.setColorSpace(QColorSpace(QColorSpace::Bt2020));
+        }
+#endif
     }
 
     // *** Set the metadata
@@ -756,7 +887,16 @@ bool RAWHandler::read(QImage *image)
     }
 
     QImage img;
-    if (!LoadRAW(this, img)) {
+    auto ok = false;
+    if (m_quality == 0) {
+        ok = LoadTHUMB(this, img);
+        if (!ok && !dev->isSequential())
+            dev->seek(m_startPos);
+    }
+    if (!ok) {
+        ok = LoadRAW(this, img);
+    }
+    if (!ok) {
         return false;
     }
 
@@ -806,12 +946,10 @@ QVariant RAWHandler::option(ImageOption option) const
         rawProcessor->imgdata.rawparams.shot_select = currentImageNumber();
 #endif
         if (rawProcessor->open_datastream(&stream) == LIBRAW_SUCCESS) {
-            if (rawProcessor->unpack() == LIBRAW_SUCCESS) {
-                auto w = libraw_get_iwidth(&rawProcessor->imgdata);
-                auto h = libraw_get_iheight(&rawProcessor->imgdata);
-                // flip & 4: taken from LibRaw code
-                v = (rawProcessor->imgdata.sizes.flip & 4) ? QSize(h, w) : QSize(w, h);
-            }
+            auto w = libraw_get_iwidth(&rawProcessor->imgdata);
+            auto h = libraw_get_iheight(&rawProcessor->imgdata);
+            // flip & 4: taken from LibRaw code
+            v = (rawProcessor->imgdata.sizes.flip & 4) ? QSize(h, w) : QSize(w, h);
         }
         d->rollbackTransaction();
     }
