@@ -20,15 +20,16 @@
  */
 
 #include "rgb_p.h"
+#include "scanlineconverter_p.h"
 #include "util_p.h"
 
 #include <cstring>
 
-#include <QList>
-#include <QMap>
-
+#include <QColorSpace>
 #include <QDebug>
 #include <QImage>
+#include <QList>
+#include <QMap>
 
 class RLEData : public QList<uchar>
 {
@@ -133,8 +134,8 @@ private:
 
     bool writeHeader();
     bool writeRle();
-    bool writeVerbatim(const QImage &);
-    bool scanData(const QImage &);
+    bool writeVerbatim(const QImage &, const QImage::Format&, const QColorSpace&);
+    bool scanData(const QImage &, const QImage::Format&, const QColorSpace&);
     uint compact(uchar *, uchar *);
     uchar intensity(uchar);
 };
@@ -225,7 +226,7 @@ bool SGIImagePrivate::readData(QImage &img)
         if (!getRow(line)) {
             return false;
         }
-        c = (QRgb *)img.scanLine(_ysize - y - 1);
+        c = reinterpret_cast<QRgb *>(img.scanLine(_ysize - y - 1));
         for (x = 0; x < _xsize; x++, c++) {
             *c = qRgb(line[x], line[x], line[x]);
         }
@@ -243,7 +244,7 @@ bool SGIImagePrivate::readData(QImage &img)
             if (!getRow(line)) {
                 return false;
             }
-            c = (QRgb *)img.scanLine(_ysize - y - 1);
+            c = reinterpret_cast<QRgb *>(img.scanLine(_ysize - y - 1));
             for (x = 0; x < _xsize; x++, c++) {
                 *c = qRgb(qRed(*c), line[x], line[x]);
             }
@@ -256,7 +257,7 @@ bool SGIImagePrivate::readData(QImage &img)
             if (!getRow(line)) {
                 return false;
             }
-            c = (QRgb *)img.scanLine(_ysize - y - 1);
+            c = reinterpret_cast<QRgb *>(img.scanLine(_ysize - y - 1));
             for (x = 0; x < _xsize; x++, c++) {
                 *c = qRgb(qRed(*c), qGreen(*c), line[x]);
             }
@@ -274,7 +275,7 @@ bool SGIImagePrivate::readData(QImage &img)
         if (!getRow(line)) {
             return false;
         }
-        c = (QRgb *)img.scanLine(_ysize - y - 1);
+        c = reinterpret_cast<QRgb *>(img.scanLine(_ysize - y - 1));
         for (x = 0; x < _xsize; x++, c++) {
             *c = qRgba(qRed(*c), qGreen(*c), qBlue(*c), line[x]);
         }
@@ -316,6 +317,9 @@ bool SGIImagePrivate::readImage(QImage &img)
         for (l = 0; !_stream.atEnd() && l < _numrows; l++) {
             _stream >> _starttab[l];
             _starttab[l] -= 512 + _numrows * 2 * sizeof(quint32);
+            if (_stream.status() != QDataStream::Ok) {
+                return false;
+            }
         }
         for (; l < _numrows; l++) {
             _starttab[l] = 0;
@@ -324,6 +328,9 @@ bool SGIImagePrivate::readImage(QImage &img)
         _lengthtab = new quint32[_numrows];
         for (l = 0; !_stream.atEnd() && l < _numrows; l++) {
             _stream >> _lengthtab[l];
+            if (_stream.status() != QDataStream::Ok) {
+                return false;
+            }
         }
     }
 
@@ -451,7 +458,7 @@ uint SGIImagePrivate::compact(uchar *d, uchar *s)
     return dest - d;
 }
 
-bool SGIImagePrivate::scanData(const QImage &img)
+bool SGIImagePrivate::scanData(const QImage &img, const QImage::Format &tfmt, const QColorSpace &tcs)
 {
     quint32 *start = _starttab;
     QByteArray lineguard(_xsize * 2, 0);
@@ -463,6 +470,8 @@ bool SGIImagePrivate::scanData(const QImage &img)
     unsigned y;
     uint len;
 
+    ScanLineConverter scl(tfmt);
+    scl.setTargetColorSpace(tcs);
     for (y = 0; y < _ysize; y++) {
         const int yPos = _ysize - y - 1; // scanline doesn't do any sanity checking
         if (yPos >= img.height()) {
@@ -470,7 +479,7 @@ bool SGIImagePrivate::scanData(const QImage &img)
             return false;
         }
 
-        c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
+        c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, yPos));
 
         for (x = 0; x < _xsize; x++) {
             buf[x] = intensity(qRed(*c++));
@@ -491,7 +500,7 @@ bool SGIImagePrivate::scanData(const QImage &img)
                 return false;
             }
 
-            c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
+            c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, yPos));
             for (x = 0; x < _xsize; x++) {
                 buf[x] = intensity(qGreen(*c++));
             }
@@ -506,7 +515,7 @@ bool SGIImagePrivate::scanData(const QImage &img)
                 return false;
             }
 
-            c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
+            c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, yPos));
             for (x = 0; x < _xsize; x++) {
                 buf[x] = intensity(qBlue(*c++));
             }
@@ -526,7 +535,7 @@ bool SGIImagePrivate::scanData(const QImage &img)
             return false;
         }
 
-        c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
+        c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, yPos));
         for (x = 0; x < _xsize; x++) {
             buf[x] = intensity(qAlpha(*c++));
         }
@@ -578,30 +587,8 @@ bool SGIImagePrivate::isSupported() const
 
 bool SGIImagePrivate::peekHeader(QIODevice *device)
 {
-    qint64 pos = 0;
-    if (!device->isSequential()) {
-        pos = device->pos();
-    }
-
-    auto ok = false;
-    QByteArray header;
-    { // datastream is destroyed before working on device
-        header = device->read(512);
-        QDataStream ds(header);
-        ok = SGIImagePrivate::readHeader(ds, this) && isValid();
-    }
-
-    if (!device->isSequential()) {
-        return device->seek(pos) && ok;
-    }
-
-    // sequential device undo
-    auto head = header.data();
-    auto readBytes = header.size();
-    while (readBytes > 0) {
-        device->ungetChar(head[readBytes-- - 1]);
-    }
-    return ok;
+    QDataStream ds(device->peek(512));
+    return SGIImagePrivate::readHeader(ds, this) && isValid();
 }
 
 QSize SGIImagePrivate::size() const
@@ -699,7 +686,7 @@ bool SGIImagePrivate::writeRle()
     return _stream.status() == QDataStream::Ok;
 }
 
-bool SGIImagePrivate::writeVerbatim(const QImage &img)
+bool SGIImagePrivate::writeVerbatim(const QImage &img, const QImage::Format &tfmt, const QColorSpace &tcs)
 {
     _rle = 0;
     if (!writeHeader()) {
@@ -710,11 +697,15 @@ bool SGIImagePrivate::writeVerbatim(const QImage &img)
     unsigned x;
     unsigned y;
 
+    ScanLineConverter scl(tfmt);
+    scl.setTargetColorSpace(tcs);
+    QByteArray ba(_xsize, char());
     for (y = 0; y < _ysize; y++) {
-        c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+        c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, _ysize - y - 1));
         for (x = 0; x < _xsize; x++) {
-            _stream << quint8(qRed(*c++));
+            ba[x] = char(qRed(*c++));
         }
+        _stream.writeRawData(ba.data(), ba.size());
     }
 
     if (_zsize == 1) {
@@ -723,17 +714,19 @@ bool SGIImagePrivate::writeVerbatim(const QImage &img)
 
     if (_zsize != 2) {
         for (y = 0; y < _ysize; y++) {
-            c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+            c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, _ysize - y - 1));
             for (x = 0; x < _xsize; x++) {
-                _stream << quint8(qGreen(*c++));
+                ba[x] = char(qGreen(*c++));
             }
+            _stream.writeRawData(ba.data(), ba.size());
         }
 
         for (y = 0; y < _ysize; y++) {
-            c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+            c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, _ysize - y - 1));
             for (x = 0; x < _xsize; x++) {
-                _stream << quint8(qBlue(*c++));
+                ba[x] = char(qBlue(*c++));
             }
+            _stream.writeRawData(ba.data(), ba.size());
         }
 
         if (_zsize == 3) {
@@ -742,10 +735,11 @@ bool SGIImagePrivate::writeVerbatim(const QImage &img)
     }
 
     for (y = 0; y < _ysize; y++) {
-        c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+        c = reinterpret_cast<const QRgb *>(scl.convertedScanLine(img, _ysize - y - 1));
         for (x = 0; x < _xsize; x++) {
-            _stream << quint8(qAlpha(*c++));
+            ba[x] = char(qAlpha(*c++));
         }
+        _stream.writeRawData(ba.data(), ba.size());
     }
 
     return _stream.status() == QDataStream::Ok;
@@ -754,30 +748,35 @@ bool SGIImagePrivate::writeVerbatim(const QImage &img)
 bool SGIImagePrivate::writeImage(const QImage &image)
 {
     //     qDebug() << "writing "; // TODO add filename
-    QImage img = image;
-    if (img.allGray()) {
+    if (image.allGray()) {
         _dim = 2, _zsize = 1;
     } else {
         _dim = 3, _zsize = 3;
     }
 
-    auto hasAlpha = img.hasAlphaChannel();
+    auto hasAlpha = image.hasAlphaChannel();
     if (hasAlpha) {
         _dim = 3, _zsize++;
     }
 
-    if (hasAlpha && img.format() != QImage::Format_ARGB32) {
-        img = img.convertToFormat(QImage::Format_ARGB32);
-    } else if (!hasAlpha && img.format() != QImage::Format_RGB32) {
-        img = img.convertToFormat(QImage::Format_RGB32);
-    }
-    if (img.isNull()) {
-        //         qDebug() << "can't convert image to depth 32";
-        return false;
+#if QT_VERSION >= QT_VERSION_CHECK(6, 8, 0)
+    auto cs = image.colorSpace();
+    auto tcs = QColorSpace();
+    auto tfmt = image.format();
+    if (cs.isValid() && cs.colorModel() == QColorSpace::ColorModel::Cmyk && tfmt == QImage::Format_CMYK8888) {
+        tcs = QColorSpace(QColorSpace::SRgb);
+        tfmt = QImage::Format_RGB32;
+    } else if (hasAlpha && tfmt != QImage::Format_ARGB32) {
+#else
+    if (hasAlpha && tfmt != QImage::Format_ARGB32) {
+#endif
+        tfmt = QImage::Format_ARGB32;
+    } else if (!hasAlpha && tfmt != QImage::Format_RGB32) {
+        tfmt = QImage::Format_RGB32;
     }
 
-    const int w = img.width();
-    const int h = img.height();
+    const int w = image.width();
+    const int h = image.height();
 
     if (w > 65535 || h > 65535) {
         return false;
@@ -794,7 +793,7 @@ bool SGIImagePrivate::writeImage(const QImage &image)
     _starttab = new quint32[_numrows];
     _rlemap.setBaseOffset(512 + _numrows * 2 * sizeof(quint32));
 
-    if (!scanData(img)) {
+    if (!scanData(image, tfmt, tcs)) {
         //         qDebug() << "this can't happen";
         return false;
     }
@@ -808,7 +807,7 @@ bool SGIImagePrivate::writeImage(const QImage &image)
     }
 
     if (verbatim_size <= rle_size) {
-        return writeVerbatim(img);
+        return writeVerbatim(image, tfmt, tcs);
     }
     return writeRle();
 }
