@@ -13,6 +13,11 @@
 #include <QImage>
 #include <QImageIOHandler>
 
+// Default maximum width and height for the large image plugins.
+#ifndef KIF_LARGE_IMAGE_PIXEL_LIMIT
+#define KIF_LARGE_IMAGE_PIXEL_LIMIT 300000
+#endif
+
 // Image metadata keys to use in plugins (so they are consistent)
 #define META_KEY_ALTITUDE "Altitude"
 #define META_KEY_AUTHOR "Author"
@@ -20,10 +25,12 @@
 #define META_KEY_COPYRIGHT "Copyright"
 #define META_KEY_CREATIONDATE "CreationDate"
 #define META_KEY_DESCRIPTION "Description"
+#define META_KEY_DIRECTION "Direction"
 #define META_KEY_DOCUMENTNAME "DocumentName"
 #define META_KEY_HOSTCOMPUTER "HostComputer"
 #define META_KEY_LATITUDE "Latitude"
 #define META_KEY_LONGITUDE "Longitude"
+#define META_KEY_MODIFICATIONDATE "ModificationDate"
 #define META_KEY_OWNER "Owner"
 #define META_KEY_SOFTWARE "Software"
 #define META_KEY_TITLE "Title"
@@ -59,4 +66,13 @@ inline QImage imageAlloc(qint32 width, qint32 height, const QImage::Format &form
     return imageAlloc(QSize(width, height), format);
 }
 
+inline double qRoundOrZero(double d)
+{
+    // If the value d is outside the range of int, the behavior is undefined.
+    if (d > std::numeric_limits<int>::max()) {
+        return 0;
+    }
+    return qRound(d);
+}
+
 #endif // UTIL_P_H
